import os
import socket

from delphai_utils.config import get_config
from motor.motor_asyncio import AsyncIOMotorClient
from pymongo import MongoClient

DB_CONN_BY_PID_AND_DBNAME = {}  # (process id, db_name) -> DB_CONN

db_connection_string = get_config("database.connection_string")
default_database_name = get_config("database.name") or "main"

MAX_POOL_SIZE = 8
MAX_IDLE_TIMEOUT = 60

mongo_parameters = dict(
    host=db_connection_string,
    appname=socket.gethostname(),
    connect=False,
    maxPoolSize=MAX_POOL_SIZE,
    maxIdleTimeMS=MAX_IDLE_TIMEOUT * 1000,
)

db_client = AsyncIOMotorClient(**mongo_parameters)
db = db_client[default_database_name]
db_sync_client = MongoClient(**mongo_parameters)
db_sync = db_sync_client[default_database_name]


def get_own_db_connection(db_name: str = None):
    """
    Creates neq connection to database.
    :param db_name: use this parameter only if DB name differs from `default_database_name`
    :return: database connection
    """
    pid = os.getpid()
    if db_name is None:
        req_db_name = default_database_name

    if (pid, req_db_name) in DB_CONN_BY_PID_AND_DBNAME:
        return DB_CONN_BY_PID_AND_DBNAME[(pid, req_db_name)]
    client = MongoClient(**mongo_parameters)

    res = client[req_db_name]
    DB_CONN_BY_PID_AND_DBNAME[(pid, req_db_name)] = res
    return res



from elasticsearch import AsyncElasticsearch

from delphai_utils.config import get_config

es_host = get_config("elasticsearch.host")
es_index = get_config("elasticsearch.index")
es_username = get_config("elasticsearch.username")
es_password = get_config("elasticsearch.password")
es_port = get_config("elasticsearch.port")
es_address = es_host
if not es_address.startswith("https://"):
    es_address = f"http://{es_host}:{es_port or 80}"

es = AsyncElasticsearch(
    hosts=[es_address], http_compress=True, http_auth=(es_username, es_password)
)



import asyncio
import collections.abc
import functools
import logging
from typing import Iterable, Optional, Union

from google.protobuf.descriptor import FileDescriptor
from grpc import Server, StatusCode, aio
from grpc.aio import AioRpcError
from grpc_health.v1 import health, health_pb2_grpc
from grpc_health.v1.health_pb2 import _HEALTH
from grpc_reflection.v1alpha import reflection
from prometheus_client import start_http_server

from delphai_utils.gateway import start_gateway
from delphai_utils.interceptors.authentication import AuthenticationInterceptor
from delphai_utils.interceptors.metrics import MetricsInterceptor
from delphai_utils.keycloak import update_public_keys
from delphai_utils.utils import find_free_port

logger = logging.getLogger(__name__)


GRPC_MAX_MESSAGE_LENGTH = 512 * 1024 * 1024

GRPC_OPTIONS = [
    ("grpc.max_send_message_length", GRPC_MAX_MESSAGE_LENGTH),
    ("grpc.max_receive_message_length", GRPC_MAX_MESSAGE_LENGTH),
]

shutdown_event = asyncio.Event()


def grpc_error(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        context = args[2]
        try:
            result = await func(*args, **kwargs)
            return result
        except AioRpcError as ex:
            await context.abort(ex.code(), ex.details())
        except Exception as ex:
            await context.abort(StatusCode.INTERNAL, str(ex))

    return wrapper


def create_grpc_server(
    descriptors: Union[FileDescriptor, Iterable[FileDescriptor]],
    server: Optional[aio.Server] = None,
) -> aio.Server:
    """Configures a grpc server based on one or multiple proto file descriptors.

    A existing grpc server can be passed as configuration base. If not the
    default configured grpc.aio.Server will be used.
    """
    if not isinstance(descriptors, collections.abc.Iterable):
        descriptors = [descriptors]

    server = server or aio.server(
        options=GRPC_OPTIONS,
        interceptors=(
            AuthenticationInterceptor(),
            MetricsInterceptor(),
        ),
    )

    health_servicer = health.HealthServicer(experimental_non_blocking=True)
    health_pb2_grpc.add_HealthServicer_to_server(health_servicer, server)

    service_names = [
        service.full_name
        for descriptor in descriptors
        for service in descriptor.services_by_name.values()
    ] + [
        _HEALTH.full_name,
        reflection.SERVICE_NAME,
    ]

    reflection.enable_server_reflection(service_names, server)
    server.descriptors = descriptors
    return server


def start_server(
    server: Server,
    gateway: bool = True,
    grpc_port: Optional[int] = None,
    http_port: Optional[int] = None,
    metrics_port: Optional[int] = None,
):
    """Start a grpc server including gateway and background tasks."""
    logger.info("starting grpc server...")

    if not grpc_port:
        grpc_port = find_free_port(8080)

    server.add_insecure_port(f"[::]:{grpc_port}")
    loop = asyncio.get_event_loop()
    loop.run_until_complete(server.start())
    logger.info(f"started grpc server on port {grpc_port}")

    if not metrics_port:
        metrics_port = find_free_port(9191)
    start_http_server(metrics_port)
    logger.info(f"started metrics server on port {metrics_port}")

    tasks = [loop.create_task(server.wait_for_termination())]
    if gateway:
        tasks.append(start_gateway(server, grpc_port, http_port))

    tasks.append(loop.create_task(update_public_keys()))

    try:
        loop.run_until_complete(asyncio.gather(*tasks))
    except KeyboardInterrupt:
        logger.info("stopped server (keyboard interrupt)")



from delphai_utils.validator import validate as _validate, ValidationFailed
from grpc import StatusCode


async def validate(descriptor, value, context):
    try:
        validated = _validate(descriptor)(value)
        return
    except ValidationFailed as ex:
        args = list(ex.args)
        errors = list(map(lambda err: err.replace("p.", ""), args))
        await context.abort(StatusCode.INVALID_ARGUMENT, ", ".join(errors))



from typing import List, Union
from delphai_utils.users import get_user
from grpc import StatusCode
from grpc.aio import ServicerContext
import asyncio
from functools import wraps


def get_roles(provided_roles: Union[List[str], ServicerContext]):
    """
    Utility function to retrieve roles from grpc context.
    :param Union[List[str], grpc.aio.ServicerContext] provided_roles:
      a list of roles or the grpc context passed to class methods
    :return list of roles
    :rtype List[str]
    """

    assert isinstance(provided_roles, object) or type(provided_roles) == list
    if type(provided_roles) is not list:
        try:
            user = get_user(provided_roles)
            return user.get("roles") or []
        except KeyError:
            return []
    else:
        return provided_roles


def is_authorized(
    provided_roles: Union[List[str], ServicerContext],
    required_roles: List[str],
    decission_logic: str = "all",
    abort: bool = False,
):
    """
    Makes a decision if user is authorized based on intersection of roles.
    :param Union[List[str], grpc.aio.ServicerContext] provided_roles:
      a list of roles or the grpc context passed to class methods
    :param List[str] required_roles: roles the request must pass
    :param str decission_logic: one of 'all' or 'some'. Default: 'all'
    :param bool abort: if set to true grpc call will abort with PERMISSION_DENIED. Default: False
    :rtype: bool
    """

    assert len(required_roles) != 0
    assert type(required_roles) == list
    assert any([allowed == decission_logic for allowed in ["all", "some"]])
    resolved_roles = get_roles(provided_roles)

    required_roles = set(required_roles)  # in case same role is passed mulitple times
    role_intersection = required_roles.intersection(resolved_roles)
    all_intersects = len(role_intersection) == len(required_roles)
    some_intersects = len(role_intersection) > 0
    _is_authorized = (decission_logic == "all" and all_intersects) or (
        decission_logic == "some" and some_intersects
    )
    is_grpc_context = type(provided_roles) is not list
    if abort and not is_authorized and is_grpc_context:
        details = f"Provided roles {resolved_roles} don't intersect with required roles {required_roles}."
        provided_roles.abort(StatusCode.PERMISSION_DENIED, details)
    return _is_authorized


def authorize(required_roles: List[str], decission_logic: str = "all"):
    """
    Decorator for grpc endpoints to restrict role-based access.
    :param List[str] required_roles: list of strings which roles are required
    :param str decission_logic: one of 'all' or 'some'. Default: 'all'
    :return: wraped function
    :rtype: func
    """

    def wrap(func):
        if not asyncio.iscoroutinefunction(func):

            @wraps(func)
            def wrapped_func(self, request, context, *args, **kwargs):
                if is_authorized(context, required_roles, decission_logic):
                    return func(self, request, context, *args, **kwargs)
                else:
                    details = f"Given roles don't intersect with {required_roles} using decission logic '{decission_logic}'"
                    context.abort(StatusCode.PERMISSION_DENIED, details)

            return wrapped_func
        else:

            @wraps(func)
            async def wrapped_func(self, request, context, *args, **kwargs):
                if is_authorized(context, required_roles, decission_logic):
                    return await func(self, request, context, *args, **kwargs)
                else:
                    details = f"Given roles don't intersect with {required_roles} using decission logic '{decission_logic}'"
                    await context.abort(StatusCode.PERMISSION_DENIED, details)

            return wrapped_func

    return wrap



from delphai_utils.authorization import is_authorized

test_cases = [
    [[[], ["1"], "all"], False],
    [[[], ["1"], "some"], False],
    [[["1", "2", "3"], ["2", "3"], "all"], True],
    [[["1", "2", "3"], ["2", "4"], "all"], False],
    [[["1", "2", "3"], ["2", "4"], "some"], True],
    [[["1", "2", "3"], ["2", "3", "3", "2"], "all"], True],
]


def test_is_authorized():
    for test_case in test_cases:
        args, expected = test_case
        assert is_authorized(*args) == expected



import logging

import grpc
from grpc.aio import ServerInterceptor
from jose.exceptions import JWTError

from delphai_utils.keycloak import PublicKeyFetchError, decode_token

logger = logging.getLogger(__name__)


class AuthenticationInterceptor(ServerInterceptor):
    async def intercept_service(self, continuation, handler_call_details):
        metadata = dict(handler_call_details.invocation_metadata)
        authorization_header = metadata.get("authorization")

        if not authorization_header:
            logger.warning("authorization header not specified")

        else:
            error_message = None

            if "Bearer " not in authorization_header:
                error_message = "Authorization header has the wrong format."

            else:
                access_token = authorization_header.split("Bearer ")[1]
                try:
                    await decode_token(access_token)
                except JWTError as ex:
                    error_message = f"Error decoding the token: {ex}"
                except PublicKeyFetchError as ex:
                    error_message = f"Error fetching jwk from keycloak: {ex}"

            if error_message is not None:

                @grpc.unary_unary_rpc_method_handler
                async def abort(ignored_request, context):
                    logger.error(error_message)
                    await context.abort(grpc.StatusCode.UNAUTHENTICATED, error_message)

                return abort

        return await continuation(handler_call_details)



from typing import Dict, List, Tuple, Union
from grpc.aio import ServicerContext
from delphai_utils.keycloak import decode_token
import asyncio
import nest_asyncio

nest_asyncio.apply()


def get_user(context: ServicerContext) -> Dict:
    """
    Get x-delphai-user information.
    :param grpc.aio.ServicerContext context: context passed to grpc endpoints
    :raises: KeyError
    :return x-delphai-user dictionary.
      More information: https://wiki.delphai.dev/wiki/Authorization
    :rtype: Dict
    """

    metadata = dict(context.invocation_metadata())
    if "authorization" not in metadata:
        return {}

    authorization_header = metadata["authorization"]
    access_token = authorization_header.split("Bearer ")[1]
    decoded_access_token = asyncio.get_event_loop().run_until_complete(
        decode_token(access_token)
    )
    user = {
        "https://delphai.com/mongo_user_id": decoded_access_token.get("mongo_user_id"),
        "https://delphai.com/client_id": decoded_access_token.get("mongo_client_id"),
    }
    if "realm_access" in decoded_access_token and "roles" in decoded_access_token.get(
        "realm_access"
    ):
        roles = decoded_access_token.get("realm_access").get("roles")
        user["roles"] = roles
    if "group_membership" in decoded_access_token:
        user["groups"] = decoded_access_token["group_membership"]
    if "limited_dataset_group_name" in decoded_access_token:
        user["limited_dataset_group_name"] = decoded_access_token[
            "limited_dataset_group_name"
        ]
    if "name" in decoded_access_token:
        user["name"] = decoded_access_token["name"]
    return user


def get_groups(context: ServicerContext) -> List[str]:
    """
    Gets groups of calling identity
    :param grpc.aio.ServicerContext context: context passed to grpc endpoints
    :return raw roles passed from keycloak.
      For example this can be ["/delphai/Development"]
    :rtype List[str]
    """

    assert isinstance(context, object)
    try:
        user = get_user(context)
    except KeyError:
        return []
    return user.get("groups") or []


def get_affiliation(
    context: ServicerContext,
) -> Tuple[Union[str, None], Union[str, None]]:
    """
    Gets organization and department of user
    :param grpc.aio.ServicerContext context: context passed to grpc endpoints
    :return organization and department as a tuple or None if not affiliated
    :rtype Union[Tuple[str, None], Tuple[str, None]]
    """

    raw_groups = get_groups(context)
    if len(raw_groups) == 0:
        return None, None
    else:
        group_hirarchy = raw_groups[0].split("/")[1:3]
        if len(group_hirarchy) == 1:
            return group_hirarchy.pop(), None
        elif len(group_hirarchy) == 2:
            return tuple(group_hirarchy)
        else:
            return None, None


def get_user_and_client_ids(context: ServicerContext) -> Tuple[str, str]:
    """
    Get user_id and client_id.
    :param grpc.aio.ServicerContext context: context passed to grpc endpoints
    :return user_id and client_id.
      More information: https://wiki.delphai.dev/wiki/Authorization
    :rtype: Tuple[str, str]
    """
    try:
        user = get_user(context)
        user_id = user.get("https://delphai.com/mongo_user_id")
        client_id = user.get("https://delphai.com/client_id")
    except Exception:
        return "", ""
    return user_id, client_id



import contextlib
import functools
import logging
import time
from typing import Dict, cast

import grpc
from grpc import StatusCode
from grpc.aio import ServerInterceptor
from prometheus_client import (
    PLATFORM_COLLECTOR,
    PROCESS_COLLECTOR,
    REGISTRY,
    Counter,
    Histogram,
)
from prometheus_client.registry import CollectorRegistry

logger = logging.getLogger(__name__)

REGISTRY.unregister(PROCESS_COLLECTOR)
REGISTRY.unregister(PLATFORM_COLLECTOR)
REGISTRY.unregister(REGISTRY._names_to_collectors["python_gc_objects_collected_total"])

"""As ServicerContext.code() returns the integer value of the StatusCode we
want to get the enum value for it"""
STATUS_CODE_BY_INT_VALUE: Dict[int, grpc.StatusCode] = {
    cast(int, enum_field.value[0]): enum_field for enum_field in grpc.StatusCode
}


class MetricsInterceptor(ServerInterceptor):
    registry: CollectorRegistry
    request_count: Counter
    latency_seconds: Histogram

    def __init__(self, registry: CollectorRegistry = REGISTRY) -> None:
        self.request_count = Counter(
            "request_count",
            "Total number of RPCs started on the server.",
            ["grpc_service", "grpc_method", "grpc_code"],
            registry=registry,
        )
        self.latency_seconds = Histogram(
            "latency_seconds",
            "Histogram of response latency (seconds)",
            ["grpc_service", "grpc_method", "grpc_code"],
            registry=registry,
        )

    async def intercept_service(self, continuation, handler_call_details):
        service_name, method_name = self._split_service_method_names(
            handler_call_details
        )

        continuation_handler = await continuation(handler_call_details)

        if not service_name.startswith("grpc."):
            continuation_handler_class = type(continuation_handler)
            continuation_handler = continuation_handler_class(
                request_streaming=continuation_handler.request_streaming,
                response_streaming=continuation_handler.response_streaming,
                request_deserializer=continuation_handler.request_deserializer,
                response_serializer=continuation_handler.response_serializer,
                unary_unary=self._wrap_unary_response(
                    continuation_handler.unary_unary, service_name, method_name
                ),
                unary_stream=self._wrap_stream_response(
                    continuation_handler.unary_stream, service_name, method_name
                ),
                stream_unary=self._wrap_unary_response(
                    continuation_handler.stream_unary, service_name, method_name
                ),
                stream_stream=self._wrap_stream_response(
                    continuation_handler.stream_stream, service_name, method_name
                ),
            )

        return continuation_handler

    def _wrap_unary_response(self, original, service_name, method_name):
        if not original:
            return original

        @functools.wraps(original)
        async def behavior(request_or_iterator, servicer_context):
            with self._track_request(service_name, method_name, servicer_context):
                return await original(request_or_iterator, servicer_context)

        return behavior

    def _wrap_stream_response(self, original, service_name, method_name):
        if not original:
            return original

        @functools.wraps(original)
        async def behavior(request_or_iterator, servicer_context):
            with self._track_request(service_name, method_name, servicer_context):
                async for chunk in original(request_or_iterator, servicer_context):
                    yield chunk

        return behavior

    def _split_service_method_names(self, handler_call_details):
        """
        Infers the grpc service and method name from the handler_call_details.
        """

        # e.g. /package.ServiceName/MethodName
        parts = handler_call_details.method.split("/")
        if len(parts) < 3:
            return "", ""

        service_name, method_name = parts[1:3]
        return service_name, method_name

    @contextlib.contextmanager
    def _track_request(self, service_name, method_name, servicer_context):
        grpc_code = StatusCode.UNKNOWN
        start_time = time.perf_counter()
        try:
            yield
            grpc_code = StatusCode.OK
        except grpc.aio.AbortError as error:
            grpc_code = STATUS_CODE_BY_INT_VALUE.get(
                servicer_context.code(), StatusCode.UNKNOWN
            )
            logger.error(f"[{grpc_code}] {error}")
            raise error

        finally:
            end_time = time.perf_counter()
            elapsed = max(end_time - start_time, 0)

            logger.info(
                f"[{grpc_code}] {service_name}/{method_name} [{elapsed * 1000:.2f}ms]"
            )
            labels = {
                "grpc_service": service_name,
                "grpc_method": method_name,
                "grpc_code": grpc_code,
            }
            self.request_count.labels(**labels).inc()
            self.latency_seconds.labels(**labels).observe(elapsed)



import logging
import os
from base64 import b64decode
from time import sleep

from dotenv import load_dotenv
from kubernetes import client as kube_client
from kubernetes import config as kube_config
from kubernetes.config import ConfigException
from memoization import cached
from omegaconf import OmegaConf

logger = logging.getLogger(__name__)


env_path = os.path.abspath(".env")
load_dotenv(dotenv_path=env_path)


def __connect_to_kubernetes():
    try:
        kube_config.load_incluster_config()
    except Exception as ex:
        logger.debug(str(ex))
        kube_config.load_kube_config()


def __get_keyring_secret(name: str):
    import keyring

    service_n_user = name.split("/")
    service_name = "/".join(service_n_user[:-1])
    username = "".join(service_n_user[-1:])
    return keyring.get_password(service_name, username) or "???"


def __get_kubernetes_secret(path: str):
    try:
        __connect_to_kubernetes()
    except ConfigException:
        logger.warn(
            f'Failed to obtain a secret value "kube:{path}": no available configuration'
        )
        return "???"
    except Exception as exc:
        logger.warn(
            f'Failed to obtain a secret value "kube:{path}": {type(exc).__name__}'
        )
        return "???"

    split = path.split("/")
    namespace = "default"
    secret_name = split[0]
    secret_key = split[1]
    if len(split) > 2:
        namespace = split[0]
        secret_name = split[1]
        secret_key = split[2]
    v1 = kube_client.CoreV1Api()

    try:
        secret = v1.read_namespaced_secret(secret_name, namespace).data[secret_key]
    except Exception as exc:
        logger.warn(
            f'Failed to obtain a secret value "kube:{path}": {type(exc).__name__}'
        )
        return "???"

    secret = b64decode(secret).decode("utf-8")
    return secret


OmegaConf.register_new_resolver("kube", __get_kubernetes_secret)
OmegaConf.register_new_resolver("keyring", __get_keyring_secret)

CURRENT_CONFIG = None
CONFIG_IS_LOADING = False


def __load_config(config_dir: str = "./config"):
    global CURRENT_CONFIG, CONFIG_IS_LOADING
    while CONFIG_IS_LOADING:
        sleep(0.05)
    if CURRENT_CONFIG is not None:
        return CURRENT_CONFIG
    while CONFIG_IS_LOADING:
        sleep(0.05)
    CONFIG_IS_LOADING = True
    try:
        if "DELPHAI_ENVIRONMENT" not in os.environ:
            raise Exception("DELPHAI_ENVIRONMENT is not defined")
        default_config = OmegaConf.load(f"{config_dir}/default.yml")
        delphai_environment = os.environ.get("DELPHAI_ENVIRONMENT")
        if os.path.isfile(f"{config_dir}/{delphai_environment}.yml"):
            delphai_env_config = OmegaConf.load(
                f"{config_dir}/{delphai_environment}.yml"
            )
        else:
            delphai_env_config = OmegaConf.create()
        CURRENT_CONFIG = OmegaConf.merge(default_config, delphai_env_config)
        OmegaConf.set_readonly(CURRENT_CONFIG, True)
        return CURRENT_CONFIG
    finally:
        CONFIG_IS_LOADING = False


@cached
def get_config(path: str = "", config_dir: str = "./config"):
    config = __load_config(config_dir=config_dir)
    if path is None:
        return config
    selected = OmegaConf.select(config, path)
    if OmegaConf.is_config(selected):
        return OmegaConf.to_container(selected, resolve=True)
    else:
        return selected



import logging as _logging
import warnings

import coloredlogs

from delphai_utils.config import get_config


LOGGING_LEVEL = "INFO"

LOGGING_FORMAT = "[%(asctime)s.%(msecs)03d] [%(levelname)s] [%(name)s] %(message)s"

LOGGERS_RESET = [
    "uvicorn",
    "uvicorn.error",
    "uvicorn.access",
]

LOGGERS_LEVELS = {
    "aiohttp": "ERROR",
    "aiokafka": "ERROR",
    "azure": "ERROR",
    "elasticsearch": "ERROR",
    "faust": "ERROR",
    "httpx": "WARNING",
    "mode": "ERROR",
    "urllib3": "ERROR",
}


def configure_logging(*, loggers_levels={}):
    logging_config = get_config("logging") or {}

    coloredlogs.install(
        level=_logging.NOTSET, fmt=logging_config.get("format", LOGGING_FORMAT)
    )

    for logger_name in LOGGERS_RESET:
        logger_object = _logging.getLogger(logger_name)
        logger_object.propagate = True
        logger_object.setLevel(_logging.NOTSET)
        logger_object.handlers.clear()

    _logging.root.setLevel(logging_config.get("level", LOGGING_LEVEL))

    loggers_levels = dict(LOGGERS_LEVELS, **(loggers_levels or {}))
    for logger_name, logger_level in loggers_levels.items():
        _logging.getLogger(logger_name).setLevel(logger_level)

    for logger_name, logger_config in logging_config.get("loggers", {}).items():
        if "level" in logger_config:
            _logging.getLogger(logger_name).setLevel(logger_config["level"])


def error_with_traceback():
    """
    Writes to log an occured exception with a last line of traceback
    """

    warnings.warn(
        'Use `logger.exception("Your message")` instead of:',
        RuntimeWarning,
        stacklevel=2,
    )
    _logging.exception("Error")


class _LoggingWrapper:
    def __getattr__(self, name):
        if "warning_showed" not in self.__dict__:
            self.warning_showed = True

            if not _logging.root.handlers:
                warnings.warn(
                    "You need to initialize logging by calling `delphai_utils.logging.configure_logging()`",
                    RuntimeWarning,
                    stacklevel=2,
                )

        warnings.warn(
            "Use internal logging ( https://wiki.delphai.dev/wiki/Logging ) instead of:",
            RuntimeWarning,
            stacklevel=2,
        )
        return getattr(_logging, name)


logging = _LoggingWrapper()



import asyncio
import sys


def aiokafka_patch():
    async def ensure_active_group(self, subscription, prev_assignment):
        # due to a race condition between the initial metadata
        # fetch and the initial rebalance, we need to ensure that
        # the metadata is fresh before joining initially. This
        # ensures that we have matched the pattern against the
        # cluster's topics at least once before joining.
        # Also the rebalance can be issued by another node, that
        # discovered a new topic, which is still unknown to this
        # one.
        if self._subscription.subscribed_pattern:
            await self._client.force_metadata_update()
            if not subscription.active:
                return None

        if not self._performed_join_prepare:
            # NOTE: We pass the previously used assignment here.
            await self._on_join_prepare(prev_assignment)
            self._performed_join_prepare = True

        # NOTE: we did not stop heartbeat task before to keep the
        # member alive during the callback, as it can commit offsets.
        # See the ``RebalanceInProgressError`` case in heartbeat
        # handling.
        await self._stop_heartbeat_task()

        # We will not attempt rejoin if there is no activity on consumer
        idle_time = self._subscription.fetcher_idle_time
        if prev_assignment is not None and idle_time >= self._max_poll_interval:
            await asyncio.sleep(self._retry_backoff_ms / 1000)
            # This if condition always triggered and resulted in the agent not being able to
            # reconnect because it kept waiting and returned None
            # return None
        # We will only try to perform the rejoin once. If it fails,
        # we will spin this loop another time, checking for coordinator
        # and subscription changes.
        # NOTE: We do re-join in sync. The group rebalance will fail on
        # subscription change and coordinator failure by itself and
        # this way we don't need to worry about racing or cancellation
        # issues that could occur if re-join were to be a task.
        success = await self._do_rejoin_group(subscription)
        if success:
            self._performed_join_prepare = False
            self._start_heartbeat_task()
            return subscription.assignment
        return None

    from aiokafka.consumer.group_coordinator import GroupCoordinator

    GroupCoordinator.ensure_active_group = ensure_active_group


def mode_traceback_patch():
    from mode.utils import tracebacks

    original_function = tracebacks.print_coro_stack

    def print_coro_stack(*args, file=sys.stderr, **kwargs):
        try:
            original_function(*args, file=file, **kwargs)
        except RuntimeError as e:
            print(repr(e), file=file)

    from mode.utils import tracebacks

    tracebacks.print_coro_stack = print_coro_stack



def clean_url(url, keep_www=False):
    """
    Format and clean and url to be saved or checked.
    Args:
        url: url to be formatted
        keep_www: keep the 'www' part of the url
    Returns: formatted url
    """

    url = url.strip()
    url = url.replace("https://", "").replace("http://", "").rstrip("/")
    if not keep_www:
        url = url.replace("www.", "")
    split_url = url.split("/")
    split_url[0] = split_url[0].lower()
    return "/".join(split_url)



import asyncio
import logging
from typing import Dict, List

import httpx
from jose import jwt

from delphai_utils.config import get_config

logger = logging.getLogger(__name__)


public_keys = {}

# endpoint for getting all enabled JWKs
base_url = get_config("keycloak.base_url") or "https://auth.delphai.com"
url = f"{base_url}/auth/realms/delphai/protocol/openid-connect/certs"


async def decode_token(access_token: str) -> str:
    decode_args = {"audience": "delphai-gateway", "options": {"leeway": 10}}
    try:
        result = jwt.decode(access_token, public_keys, **decode_args)
        return result
    except Exception:
        _public_keys = await __async_fetch_keys()
        return jwt.decode(access_token, _public_keys, **decode_args)


class PublicKeyFetchError(Exception):
    pass


async def __async_fetch_keys() -> List[Dict]:
    global public_keys
    async with httpx.AsyncClient() as client:
        response = await client.get(url)
        if response.status_code != 200:
            print("failed")
            raise PublicKeyFetchError(response.text)
        else:
            result = response.json()
            public_keys = result
            return result


async def update_public_keys(interval: int = 60 * 10):  # updates every 10 minutes
    while True:
        try:
            await __async_fetch_keys()
        except PublicKeyFetchError:
            logger.exception("Error fetching jwk from keycloak")
        await asyncio.sleep(interval)



import re
from functools import lru_cache
from validate_email import validate_email
import ipaddress

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse
import uuid
import struct
from jinja2 import Template
import time
import sys

printer = ""

# Well known regex mapping.
regex_map = {
    "UNKNOWN": "",
    "HTTP_HEADER_NAME": r"^:?[0-9a-zA-Z!#$%&\'*+-.^_|~\x60]+$",
    "HTTP_HEADER_VALUE": r"^[^\u0000-\u0008\u000A-\u001F\u007F]*$",
    "HEADER_STRING": r"^[^\u0000\u000A\u000D]*$",
}


class ValidatingMessage(object):
    """Wrap a proto message to cache validate functions with the message class name.

    A validate function is defined per message class in protoc-gen-validate,
     so we can reuse an already generated function for the same message class.
    """

    def __init__(self, proto_message):
        self.DESCRIPTOR = proto_message.DESCRIPTOR

    def __hash__(self):
        return hash(self.DESCRIPTOR.full_name)

    def __eq__(self, other):
        if isinstance(other, ValidatingMessage):
            return self.DESCRIPTOR.full_name == other.DESCRIPTOR.full_name
        else:
            return False


def validate(proto_message):
    return _validate_inner(ValidatingMessage(proto_message))


# Cache generated functions to avoid the performance issue caused by repeated proto messages,
#   which generate the same functions repeatedly.
@lru_cache()
def _validate_inner(proto_message):
    func = file_template(proto_message)
    global printer
    printer += func + "\n"
    exec(func)
    try:
        return generate_validate
    except NameError:
        return locals()["generate_validate"]


def print_validate(proto_message):
    return "".join([s for s in printer.splitlines(True) if s.strip()])


def has_validate(field):
    if field.GetOptions() is None:
        return False
    for option_descriptor, option_value in field.GetOptions().ListFields():
        if option_descriptor.full_name == "validate.rules":
            return True
    return False


def byte_len(s):
    try:
        return len(s.encode("utf-8"))
    except Exception:
        return len(s)


def _validateHostName(host):
    if not host:
        return False
    if len(host) > 253:
        return False

    if host[-1] == ".":
        host = host[:-1]

    for part in host.split("."):
        if len(part) == 0 or len(part) > 63:
            return False

        # Host names cannot begin or end with hyphens
        if part[0] == "-" or part[-1] == "-":
            return False
        for r in part:
            if (
                (r < "A" or r > "Z")
                and (r < "a" or r > "z")
                and (r < "0" or r > "9")
                and r != "-"
            ):
                return False
    return True


def _validateEmail(addr):
    if "<" in addr and ">" in addr:
        addr = addr.split("<")[1].split(">")[0]

    if not validate_email(addr):
        return False

    if len(addr) > 254:
        return False

    parts = addr.split("@")
    if len(parts[0]) > 64:
        return False
    return _validateHostName(parts[1])


def _has_field(message_pb, property_name):
    # NOTE: As of proto3, HasField() only works for message fields, not for
    #       singular (non-message) fields. First try to use HasField and
    #       if it fails (with a ValueError) we manually consult the fields.
    try:
        return message_pb.HasField(property_name)
    except Exception:
        all_fields = set([field.name for field in message_pb.DESCRIPTOR.fields])
        return property_name in all_fields


def const_template(option_value, name):
    const_tmpl = """{%- if str(o.string) and o.string.HasField('const') -%}
    if {{ name }} != \"{{ o.string['const'] }}\":
        raise ValidationFailed(\"{{ name }} not equal to {{ o.string['const'] }}\")
    {%- elif str(o.bool) and o.bool['const'] != "" -%}
    if {{ name }} != {{ o.bool['const'] }}:
        raise ValidationFailed(\"{{ name }} not equal to {{ o.bool['const'] }}\")
    {%- elif str(o.enum) and o.enum['const'] -%}
    if {{ name }} != {{ o.enum['const'] }}:
        raise ValidationFailed(\"{{ name }} not equal to {{ o.enum['const'] }}\")
    {%- elif str(o.bytes) and o.bytes.HasField('const') -%}
        {% if sys.version_info[0] >= 3 %}
    if {{ name }} != {{ o.bytes['const'] }}:
        raise ValidationFailed(\"{{ name }} not equal to {{ o.bytes['const'] }}\")
        {% else %}
    if {{ name }} != b\"{{ o.bytes['const'].encode('string_escape') }}\":
        raise ValidationFailed(\"{{ name }} not equal to {{ o.bytes['const'].encode('string_escape') }}\")
        {% endif %}
    {%- endif -%}
    """
    return Template(const_tmpl).render(sys=sys, o=option_value, name=name, str=str)


def in_template(value, name):
    in_tmpl = """
    {%- if value['in'] %}
    if {{ name }} not in {{ value['in'] }}:
        raise ValidationFailed(\"{{ name }} not in {{ value['in'] }}\")
    {%- endif -%}
    {%- if value['not_in'] %}
    if {{ name }} in {{ value['not_in'] }}:
        raise ValidationFailed(\"{{ name }} in {{ value['not_in'] }}\")
    {%- endif -%}
    """
    return Template(in_tmpl).render(value=value, name=name)


def string_template(option_value, name):
    if option_value.string.well_known_regex:
        known_regex_type = option_value.string.DESCRIPTOR.fields_by_name[
            "well_known_regex"
        ].enum_type
        regex_value = option_value.string.well_known_regex
        regex_name = known_regex_type.values_by_number[regex_value].name
        if (
            regex_name in ["HTTP_HEADER_NAME", "HTTP_HEADER_VALUE"]
            and not option_value.string.strict
        ):
            option_value.string.pattern = regex_map["HEADER_STRING"]
        else:
            option_value.string.pattern = regex_map[regex_name]
    str_templ = """
    {{ const_template(o, name) -}}
    {{ in_template(o.string, name) -}}
    {%- set s = o.string -%}
    {%- if s['len'] %}
    if len({{ name }}) != {{ s['len'] }}:
        raise ValidationFailed(\"{{ name }} length does not equal {{ s['len'] }}\")
    {%- endif -%}
    {%- if s['min_len'] %}
    if len({{ name }}) < {{ s['min_len'] }}:
        raise ValidationFailed(\"{{ name }} length is less than {{ s['min_len'] }}\")
    {%- endif -%}
    {%- if s['max_len'] %}
    if len({{ name }}) > {{ s['max_len'] }}:
        raise ValidationFailed(\"{{ name }} length is more than {{ s['max_len'] }}\")
    {%- endif -%}
    {%- if s['len_bytes'] %}
    if byte_len({{ name }}) != {{ s['len_bytes'] }}:
        raise ValidationFailed(\"{{ name }} length does not equal {{ s['len_bytes'] }}\")
    {%- endif -%}
    {%- if s['min_bytes'] %}
    if byte_len({{ name }}) < {{ s['min_bytes'] }}:
        raise ValidationFailed(\"{{ name }} length is less than {{ s['min_bytes'] }}\")
    {%- endif -%}
    {%- if s['max_bytes'] %}
    if byte_len({{ name }}) > {{ s['max_bytes'] }}:
        raise ValidationFailed(\"{{ name }} length is greater than {{ s['max_bytes'] }}\")
    {%- endif -%}
    {%- if s['pattern'] %}
    if re.search(r\'{{ s['pattern'] }}\', {{ name }}) is None:
        raise ValidationFailed(\"{{ name }} pattern does not match {{ s['pattern'] }}\")
    {%- endif -%}
    {%- if s['prefix'] %}
    if not {{ name }}.startswith(\"{{ s['prefix'] }}\"):
        raise ValidationFailed(\"{{ name }} does not start with prefix {{ s['prefix'] }}\")
    {%- endif -%}
    {%- if s['suffix'] %}
    if not {{ name }}.endswith(\"{{ s['suffix'] }}\"):
        raise ValidationFailed(\"{{ name }} does not end with suffix {{ s['suffix'] }}\")
    {%- endif -%}
    {%- if s['contains'] %}
    if not \"{{ s['contains'] }}\" in {{ name }}:
        raise ValidationFailed(\"{{ name }} does not contain {{ s['contains'] }}\")
    {%- endif -%}
    {%- if s['not_contains'] %}
    if \"{{ s['not_contains'] }}\" in {{ name }}:
        raise ValidationFailed(\"{{ name }} contains {{ s['not_contains'] }}\")
    {%- endif -%}
    {%- if s['email'] %}
    if not _validateEmail({{ name }}):
        raise ValidationFailed(\"{{ name }} is not a valid email\")
    {%- endif -%}
    {%- if s['hostname'] %}
    if not _validateHostName({{ name }}):
        raise ValidationFailed(\"{{ name }} is not a valid email\")
    {%- endif -%}
    {%- if s['address'] %}
    try:
        ipaddress.ip_address({{ name }})
    except ValueError:
        if not _validateHostName({{ name }}):
            raise ValidationFailed(\"{{ name }} is not a valid address\")
    {%- endif -%}
    {%- if s['ip'] %}
    try:
        ipaddress.ip_address({{ name }})
    except ValueError:
        raise ValidationFailed(\"{{ name }} is not a valid ip\")
    {%- endif -%}
    {%- if s['ipv4'] %}
    try:
        ipaddress.IPv4Address({{ name }})
    except ValueError:
        raise ValidationFailed(\"{{ name }} is not a valid ipv4\")
    {%- endif -%}
    {%- if s['ipv6'] %}
    try:
        ipaddress.IPv6Address({{ name }})
    except ValueError:
        raise ValidationFailed(\"{{ name }} is not a valid ipv6\")
    {%- endif %}
    {%- if s['uri'] %}
    url = urlparse.urlparse({{ name }})
    if not all([url.scheme, url.netloc, url.path]):
        raise ValidationFailed(\"{{ name }} is not a valid uri\")
    {%- endif %}
    {%- if s['uri_ref'] %}
    url = urlparse.urlparse({{ name }})
    if not all([url.scheme, url.path]) and url.fragment:
        raise ValidationFailed(\"{{ name }} is not a valid uri ref\")
    {%- endif -%}
    {%- if s['uuid'] %}
    try:
        uuid.UUID({{ name }})
    except ValueError:
        raise ValidationFailed(\"{{ name }} is not a valid UUID\")
    {%- endif -%}
    """
    return Template(str_templ).render(
        o=option_value,
        name=name,
        const_template=const_template,
        in_template=in_template,
    )


def required_template(value, name):
    req_tmpl = """{%- if value['required'] -%}
    if not _has_field(p, \"{{ name.split('.')[-1] }}\"):
        raise ValidationFailed(\"{{ name }} is required.\")
    {%- endif -%}
    """
    return Template(req_tmpl).render(value=value, name=name)


def message_template(option_value, name, repeated=False):
    message_tmpl = """{%- if m.message %}
    {{- required_template(m.message, name) }}
    {%- endif -%}
    {%- if m.message and m.message['skip'] %}
    # Skipping validation for {{ name }}
    {%- else %}
    {% if repeated %}
    if {{ name }}:
    {% else %}
    if _has_field(p, \"{{ name.split('.')[-1] }}\"):
    {% endif %}
        embedded = validate(p.{{ name }})(p.{{ name }})
        if embedded is not None:
            return embedded
    {%- endif -%}
    """
    return Template(message_tmpl).render(
        m=option_value,
        name=name,
        required_template=required_template,
        repeated=repeated,
    )


def bool_template(option_value, name):
    bool_tmpl = """
    {{ const_template(o, name) -}}
    """
    return Template(bool_tmpl).render(
        o=option_value, name=name, const_template=const_template
    )


def num_template(option_value, name, num):
    num_tmpl = """{%- if num.HasField('const') and str(o.float) == "" -%}
    if {{ name }} != {{ num['const'] }}:
        raise ValidationFailed(\"{{ name }} not equal to {{ num['const'] }}\")
    {%- endif -%}
    {%- if num.HasField('const') and str(o.float) != "" %}
    if {{ name }} != struct.unpack(\"f\", struct.pack(\"f\", ({{ num['const'] }})))[0]:
        raise ValidationFailed(\"{{ name }} not equal to {{ num['const'] }}\")
    {%- endif -%}
    {{ in_template(num, name) }}
    {%- if num.HasField('lt') %}
        {%- if num.HasField('gt') %}
            {%- if num['lt'] > num['gt'] %}
    if {{ name }} <= {{ num['gt'] }} or {{ name }} >= {{ num ['lt'] }}:
        raise ValidationFailed(\"{{ name }} is not in range {{ num['lt'], num['gt'] }}\")
            {%- else %}
    if {{ name }} >= {{ num['lt'] }} and {{ name }} <= {{ num['gt'] }}:
        raise ValidationFailed(\"{{ name }} is not in range {{ num['gt'], num['lt'] }}\")
            {%- endif -%}
        {%- elif num.HasField('gte') %}
            {%- if num['lt'] > num['gte'] %}
    if {{ name }} < {{ num['gte'] }} or {{ name }} >= {{ num ['lt'] }}:
        raise ValidationFailed(\"{{ name }} is not in range {{ num['lt'], num['gte'] }}\")
            {%- else %}
    if {{ name }} >= {{ num['lt'] }} and {{ name }} < {{ num['gte'] }}:
        raise ValidationFailed(\"{{ name }} is not in range {{ num['gte'], num['lt'] }}\")
            {%- endif -%}
        {%- else %}
    if {{ name }} >= {{ num['lt'] }}:
        raise ValidationFailed(\"{{ name }} is not lesser than {{ num['lt'] }}\")
        {%- endif -%}
    {%- elif num.HasField('lte') %}
        {%- if num.HasField('gt') %}
            {%- if num['lte'] > num['gt'] %}
    if {{ name }} <= {{ num['gt'] }} or {{ name }} > {{ num ['lte'] }}:
        raise ValidationFailed(\"{{ name }} is not in range {{ num['lte'], num['gt'] }}\")
            {%- else %}
    if {{ name }} > {{ num['lte'] }} and {{ name }} <= {{ num['gt'] }}:
        raise ValidationFailed(\"{{ name }} is not in range {{ num['gt'], num['lte'] }}\")
            {%- endif -%}
        {%- elif num.HasField('gte') %}
            {%- if num['lte'] > num['gte'] %}
    if {{ name }} < {{ num['gte'] }} or {{ name }} > {{ num ['lte'] }}:
        raise ValidationFailed(\"{{ name }} is not in range {{ num['lte'], num['gte'] }}\")
            {%- else %}
    if {{ name }} > {{ num['lte'] }} and {{ name }} < {{ num['gte'] }}:
        raise ValidationFailed(\"{{ name }} is not in range {{ num['gte'], num['lte'] }}\")
            {%- endif -%}
        {%- else %}
    if {{ name }} > {{ num['lte'] }}:
        raise ValidationFailed(\"{{ name }} is not lesser than or equal to {{ num['lte'] }}\")
        {%- endif -%}
    {%- elif num.HasField('gt') %}
    if {{ name }} <= {{ num['gt'] }}:
        raise ValidationFailed(\"{{ name }} is not greater than {{ num['gt'] }}\")
    {%- elif num.HasField('gte') %}
    if {{ name }} < {{ num['gte'] }}:
        raise ValidationFailed(\"{{ name }} is not greater than or equal to {{ num['gte'] }}\")
    {%- endif -%}
    """
    return Template(num_tmpl).render(
        o=option_value, name=name, num=num, in_template=in_template, str=str
    )


def dur_arr(dur):
    value = 0
    arr = []
    for val in dur:
        value += val.seconds
        value += 10 ** -9 * val.nanos
        arr.append(value)
        value = 0
    return arr


def dur_lit(dur):
    value = dur.seconds + (10 ** -9 * dur.nanos)
    return value


def duration_template(option_value, name, repeated=False):
    dur_tmpl = """
    {{- required_template(o.duration, name) }}
    {% if repeated %}
    if {{ name }}:
    {% else %}
    if _has_field(p, \"{{ name.split('.')[-1] }}\"):
    {% endif %}
        dur = {{ name }}.seconds + round((10**-9 * {{ name }}.nanos), 9)
        {%- set dur = o.duration -%}
        {%- if dur.HasField('lt') %}
        lt = {{ dur_lit(dur['lt']) }}
        {% endif %}
        {%- if dur.HasField('lte') %}
        lte = {{ dur_lit(dur['lte']) }}
        {% endif %}
        {%- if dur.HasField('gt') %}
        gt = {{ dur_lit(dur['gt']) }}
        {% endif %}
        {%- if dur.HasField('gte') %}
        gte = {{ dur_lit(dur['gte']) }}
        {% endif %}
        {%- if dur.HasField('const') %}
        if dur != {{ dur_lit(dur['const']) }}:
            raise ValidationFailed(\"{{ name }} is not equal to {{ dur_lit(dur['const']) }}\")
        {%- endif -%}
        {%- if dur['in'] %}
        if dur not in {{ dur_arr(dur['in']) }}:
            raise ValidationFailed(\"{{ name }} is not in {{ dur_arr(dur['in']) }}\")
        {%- endif -%}
        {%- if dur['not_in'] %}
        if dur in {{ dur_arr(dur['not_in']) }}:
            raise ValidationFailed(\"{{ name }} is not in {{ dur_arr(dur['not_in']) }}\")
        {%- endif -%}
        {%- if dur.HasField('lt') %}
            {%- if dur.HasField('gt') %}
                {%- if dur_lit(dur['lt']) > dur_lit(dur['gt']) %}
        if dur <= gt or dur >= lt:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(dur['lt']), dur_lit(dur['gt']) }}\")
                {%- else -%}
        if dur >= lt and dur <= gt:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(dur['gt']), dur_lit(dur['lt']) }}\")
                {%- endif -%}
            {%- elif dur.HasField('gte') %}
                {%- if dur_lit(dur['lt']) > dur_lit(dur['gte']) %}
        if dur < gte or dur >= lt:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(dur['lt']), dur_lit(dur['gte']) }}\")
                {%- else -%}
        if dur >= lt and dur < gte:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(dur['gte']), dur_lit(dur['lt']) }}\")
                {%- endif -%}
            {%- else -%}
        if dur >= lt:
            raise ValidationFailed(\"{{ name }} is not lesser than {{ dur_lit(dur['lt']) }}\")
            {%- endif -%}
        {%- elif dur.HasField('lte') %}
            {%- if dur.HasField('gt') %}
                {%- if dur_lit(dur['lte']) > dur_lit(dur['gt']) %}
        if dur <= gt or dur > lte:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(dur['lte']), dur_lit(dur['gt']) }}\")
                {%- else -%}
        if dur > lte and dur <= gt:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(dur['gt']), dur_lit(dur['lte']) }}\")
                {%- endif -%}
            {%- elif dur.HasField('gte') %}
                {%- if dur_lit(dur['lte']) > dur_lit(dur['gte']) %}
        if dur < gte or dur > lte:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(dur['lte']), dur_lit(dur['gte']) }}\")
                {%- else -%}
        if dur > lte and dur < gte:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(dur['gte']), dur_lit(dur['lte']) }}\")
                {%- endif -%}
            {%- else -%}
        if dur > lte:
            raise ValidationFailed(\"{{ name }} is not lesser than or equal to {{ dur_lit(dur['lte']) }}\")
            {%- endif -%}
        {%- elif dur.HasField('gt') %}
        if dur <= gt:
            raise ValidationFailed(\"{{ name }} is not greater than {{ dur_lit(dur['gt']) }}\")
        {%- elif dur.HasField('gte') %}
        if dur < gte:
            raise ValidationFailed(\"{{ name }} is not greater than or equal to {{ dur_lit(dur['gte']) }}\")
        {%- endif -%}
    """
    return Template(dur_tmpl).render(
        o=option_value,
        name=name,
        required_template=required_template,
        dur_lit=dur_lit,
        dur_arr=dur_arr,
        repeated=repeated,
    )


def timestamp_template(option_value, name, repeated=False):
    timestamp_tmpl = """
    {{- required_template(o.timestamp, name) }}
    {% if repeated %}
    if {{ name }}:
    {% else %}
    if _has_field(p, \"{{ name.split('.')[-1] }}\"):
    {% endif %}
        ts = {{ name }}.seconds + round((10**-9 * {{ name }}.nanos), 9)
        {%- set ts = o.timestamp -%}
        {%- if ts.HasField('lt') %}
        lt = {{ dur_lit(ts['lt']) }}
        {% endif -%}
        {%- if ts.HasField('lte') %}
        lte = {{ dur_lit(ts['lte']) }}
        {% endif -%}
        {%- if ts.HasField('gt') %}
        gt = {{ dur_lit(ts['gt']) }}
        {% endif -%}
        {%- if ts.HasField('gte') %}
        gte = {{ dur_lit(ts['gte']) }}
        {% endif -%}
        {%- if ts.HasField('const') %}
        if ts != {{ dur_lit(ts['const']) }}:
            raise ValidationFailed(\"{{ name }} is not equal to {{ dur_lit(ts['const']) }}\")
        {% endif %}
        {%- if ts['in'] %}
        if ts not in {{ dur_arr(ts['in']) }}:
            raise ValidationFailed(\"{{ name }} is not in {{ dur_arr(ts['in']) }}\")
        {%- endif %}
        {%- if ts['not_in'] %}
        if ts in {{ dur_arr(ts['not_in']) }}:
            raise ValidationFailed(\"{{ name }} is not in {{ dur_arr(ts['not_in']) }}\")
        {%- endif %}
        {%- if ts.HasField('lt') %}
            {%- if ts.HasField('gt') %}
                {%- if dur_lit(ts['lt']) > dur_lit(ts['gt']) %}
        if ts <= gt or ts >= lt:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(ts['lt']), dur_lit(ts['gt']) }}\")
                {%- else -%}
        if ts >= lt and ts <= gt:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(ts['gt']), dur_lit(ts['lt']) }}\")
                {%- endif -%}
            {%- elif ts.HasField('gte') %}
                {%- if dur_lit(ts['lt']) > dur_lit(ts['gte']) %}
        if ts < gte or ts >= lt:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(ts['lt']), dur_lit(ts['gte']) }}\")
                {%- else -%}
        if ts >= lt and ts < gte:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(ts['gte']), dur_lit(ts['lt']) }}\")
                {%- endif -%}
            {%- else -%}
        if ts >= lt:
            raise ValidationFailed(\"{{ name }} is not lesser than {{ dur_lit(ts['lt']) }}\")
            {%- endif -%}
        {%- elif ts.HasField('lte') %}
            {%- if ts.HasField('gt') %}
                {%- if dur_lit(ts['lte']) > dur_lit(ts['gt']) %}
        if ts <= gt or ts > lte:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(ts['lte']), dur_lit(ts['gt']) }}\")
                {%- else -%}
        if ts > lte and ts <= gt:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(ts['gt']), dur_lit(ts['lte']) }}\")
                {%- endif -%}
            {%- elif ts.HasField('gte') %}
                {%- if dur_lit(ts['lte']) > dur_lit(ts['gte']) %}
        if ts < gte or ts > lte:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(ts['lte']), dur_lit(ts['gte']) }}\")
                {%- else -%}
        if ts > lte and ts < gte:
            raise ValidationFailed(\"{{ name }} is not in range {{ dur_lit(ts['gte']), dur_lit(ts['lte']) }}\")
                {%- endif -%}
            {%- else -%}
        if ts > lte:
            raise ValidationFailed(\"{{ name }} is not lesser than or equal to {{ dur_lit(ts['lte']) }}\")
            {%- endif -%}
        {%- elif ts.HasField('gt') %}
        if ts <= gt:
            raise ValidationFailed(\"{{ name }} is not greater than {{ dur_lit(ts['gt']) }}\")
        {%- elif ts.HasField('gte') %}
        if ts < gte:
            raise ValidationFailed(\"{{ name }} is not greater than or equal to {{ dur_lit(ts['gte']) }}\")
        {%- elif ts.HasField('lt_now') %}
        now = time.time()
            {%- if ts.HasField('within') %}
        within = {{ dur_lit(ts['within']) }}
        if ts >= now or ts <= now - within:
            raise ValidationFailed(\"{{ name }} is not within range {{ dur_lit(ts['within']) }}\")
            {%- else %}
        if ts >= now:
            raise ValidationFailed(\"{{ name }} is not lesser than now\")
            {%- endif -%}
        {%- elif ts.HasField('gt_now') %}
        now = time.time()
            {%- if ts.HasField('within') %}
        within = {{ dur_lit(ts['within']) }}
        if ts <= now or ts >= now + within:
            raise ValidationFailed(\"{{ name }} is not within range {{ dur_lit(ts['within']) }}\")
            {%- else %}
        if ts <= now:
            raise ValidationFailed(\"{{ name }} is not greater than now\")
            {%- endif -%}
        {%- elif ts.HasField('within') %}
        now = time.time()
        within = {{ dur_lit(ts['within']) }}
        if ts >= now + within or ts <= now - within:
             raise ValidationFailed(\"{{ name }} is not within range {{ dur_lit(ts['within']) }}\")
        {%- endif -%}
    """
    return Template(timestamp_tmpl).render(
        o=option_value,
        name=name,
        required_template=required_template,
        dur_lit=dur_lit,
        dur_arr=dur_arr,
        repeated=repeated,
    )


def wrapper_template(option_value, name, repeated=False):
    wrapper_tmpl = """
    {% if repeated %}
    if {{ name }}:
    {% else %}
    if p.HasField(\"{{ name[2:] }}\"):
    {% endif %}
        {%- if str(option_value.float) %}
        {{- num_template(option_value, name + ".value", option_value.float)|indent(4,True) -}}
        {% endif -%}
        {%- if str(option_value.double) %}
        {{- num_template(option_value, name + ".value", option_value.double)|indent(4,True) -}}
        {% endif -%}
        {%- if str(option_value.int32) %}
        {{- num_template(option_value, name + ".value", option_value.int32)|indent(4,True) -}}
        {% endif -%}
        {%- if str(option_value.int64) %}
        {{- num_template(option_value, name + ".value", option_value.int64)|indent(4,True) -}}
        {% endif -%}
        {%- if str(option_value.uint32) %}
        {{- num_template(option_value, name + ".value", option_value.uint32)|indent(4,True) -}}
        {% endif -%}
        {%- if str(option_value.uint64) %}
        {{- num_template(option_value, name + ".value", option_value.uint64)|indent(4,True) -}}
        {% endif -%}
        {%- if str(option_value.bool) %}
        {{- bool_template(option_value, name + ".value")|indent(4,True) -}}
        {% endif -%}
        {%- if str(option_value.string) %}
        {{- string_template(option_value, name + ".value")|indent(4,True) -}}
        {% endif -%}
        {%- if str(option_value.bytes) %}
        {{- bytes_template(option_value, name + ".value")|indent(4,True) -}}
        {% endif -%}
    {%- if str(option_value.message) and option_value.message['required'] %}
    else:
        raise ValidationFailed(\"{{ name }} is required.\")
    {%- endif %}
    """
    return Template(wrapper_tmpl).render(
        option_value=option_value,
        name=name,
        str=str,
        num_template=num_template,
        bool_template=bool_template,
        string_template=string_template,
        bytes_template=bytes_template,
        repeated=repeated,
    )


def enum_values(field):
    return [x.number for x in field.enum_type.values]


def enum_template(option_value, name, field):
    enum_tmpl = """
    {{ const_template(option_value, name) -}}
    {{ in_template(option_value.enum, name) -}}
    {% if option_value.enum['defined_only'] %}
    if {{ name }} not in {{ enum_values(field) }}:
        raise ValidationFailed(\"{{ name }} is not defined\")
    {% endif %}
    """
    return Template(enum_tmpl).render(
        option_value=option_value,
        name=name,
        const_template=const_template,
        in_template=in_template,
        field=field,
        enum_values=enum_values,
    )


def any_template(option_value, name, repeated=False):
    any_tmpl = """
    {{- required_template(o, name) }}
    {%- if o['in'] %}
    {% if repeated %}
    if {{ name }}:
    {% else %}
    if _has_field(p, \"{{ name.split('.')[-1] }}\"):
    {% endif %}
        if {{ name }}.type_url not in {{ o['in'] }}:
            raise ValidationFailed(\"{{ name }} not in {{ o['in'] }}\")
    {%- endif %}
    {%- if o['not_in'] %}
    {% if repeated %}
    if {{ name }}:
    {% else %}
    if _has_field(p, \"{{ name.split('.')[-1] }}\"):
    {% endif %}
        if {{ name }}.type_url in {{ o['not_in'] }}:
            raise ValidationFailed(\"{{ name }} in {{ o['not_in'] }}\")
    {%- endif %}
    """
    return Template(any_tmpl).render(
        o=option_value.any,
        name=name,
        required_template=required_template,
        repeated=repeated,
    )


def bytes_template(option_value, name):
    bytes_tmpl = """
    {{ const_template(o, name) -}}
    {{ in_template(o.bytes, name) -}}
    {%- if b['len'] %}
    if len({{ name }}) != {{ b['len'] }}:
        raise ValidationFailed(\"{{ name }} length does not equal {{ b['len'] }}\")
    {%- endif -%}
    {%- if b['min_len'] %}
    if len({{ name }}) < {{ b['min_len'] }}:
        raise ValidationFailed(\"{{ name }} length is less than {{ b['min_len'] }}\")
    {%- endif -%}
    {%- if b['max_len'] %}
    if len({{ name }}) > {{ b['max_len'] }}:
        raise ValidationFailed(\"{{ name }} length is more than {{ b['max_len'] }}\")
    {%- endif -%}
    {%- if b['ip'] %}
    try:
        ipaddress.ip_address({{ name }})
    except ValueError:
        raise ValidationFailed(\"{{ name }} is not a valid ip\")
    {%- endif -%}
    {%- if b['ipv4'] %}
    try:
        ipaddress.IPv4Address({{ name }})
    except ValueError:
        raise ValidationFailed(\"{{ name }} is not a valid ipv4\")
    {%- endif -%}
    {%- if b['ipv6'] %}
    try:
        ipaddress.IPv6Address({{ name }})
    except ValueError:
        raise ValidationFailed(\"{{ name }} is not a valid ipv6\")
    {%- endif -%}
    {% if b['pattern'] %}
        {% if sys.version_info[0] >= 3%}
    if re.search({{ b['pattern'].encode('unicode-escape') }}, {{ name }}) is None:
        raise ValidationFailed(\"{{ name }} pattern does not match b['pattern'].encode('unicode-escape')\")
        {% else %}
    if re.search(b\"{{ b['pattern'].encode('unicode-escape') }}\", {{ name }}) is None:
        raise ValidationFailed(\"{{ name }} pattern does not match \")
        {% endif %}
    {% endif %}
    {% if b['contains'] %}
        {% if sys.version_info[0] >= 3 %}
    if not {{ b['contains'] }} in {{ name }}:
        raise ValidationFailed(\"{{ name }} does not contain {{ b['contains'] }}\")
        {% else %}
    if not b\"{{ b['contains'].encode('string_escape') }}\" in {{ name }}:
        raise ValidationFailed(\"{{ name }} does not contain \")
        {% endif %}
    {% endif %}
    {% if b['prefix'] %}
        {% if sys.version_info[0] >= 3 %}
    if not {{ name }}.startswith({{ b['prefix'] }}):
        raise ValidationFailed(\"{{ name }} does not start with prefix {{ b['prefix'] }}\")
        {% else %}
    if not {{name}}.startswith(b\"{{ b['prefix'].encode('string_escape') }}\"):
        raise ValidationFailed(\"{{ name }} does not start with prefix {{ b['prefix'].encode('string_escape') }}\")
        {% endif %}
    {% endif %}
    {% if b['suffix'] %}
        {% if sys.version_info[0] >= 3 %}
    if not {{ name }}.endswith({{ b['suffix'] }}):
        raise ValidationFailed(\"{{ name }} does not end with suffix {{ b['suffix'] }}\")
        {% else %}
    if not {{name}}.endswith(b\"{{ b['suffix'].encode('string_escape') }}\"):
        raise ValidationFailed(\"{{ name }} does not end with suffix {{ b['suffix'] }}\")
        {% endif %}
    {% endif %}
    """
    return Template(bytes_tmpl).render(
        sys=sys,
        o=option_value,
        name=name,
        const_template=const_template,
        in_template=in_template,
        b=option_value.bytes,
    )


def switcher_template(accessor, name, field, map=False):
    switcher_tmpl = """
    {%- if str(accessor.float) %}
    {{- num_template(accessor, name, accessor.float)|indent(4,True) -}}
    {%- elif str(accessor.double) %}
    {{- num_template(accessor, name, accessor.double)|indent(4,True) -}}
    {%- elif str(accessor.int32) %}
    {{- num_template(accessor, name, accessor.int32)|indent(4,True) -}}
    {%- elif str(accessor.int64) %}
    {{- num_template(accessor, name, accessor.int64)|indent(4,True) -}}
    {%- elif str(accessor.uint32) %}
    {{- num_template(accessor, name, accessor.uint32)|indent(4,True) -}}
    {%- elif str(accessor.uint64) %}
    {{- num_template(accessor, name, accessor.uint64)|indent(4,True) -}}
    {%- elif str(accessor.sint32) %}
    {{- num_template(accessor, name, accessor.sint32)|indent(4,True) -}}
    {%- elif str(accessor.sint64) %}
    {{- num_template(accessor, name, accessor.sint64)|indent(4,True) -}}
    {%- elif str(accessor.fixed32) %}
    {{- num_template(accessor, name, accessor.fixed32)|indent(4,True) -}}
    {%- elif str(accessor.fixed64) %}
    {{- num_template(accessor, name, accessor.fixed64)|indent(4,True) -}}
    {%- elif str(accessor.sfixed32) %}
    {{- num_template(accessor, name, accessor.sfixed32)|indent(4,True) -}}
    {%- elif str(accessor.sfixed64) %}
    {{- num_template(accessor, name, accessor.sfixed64)|indent(4,True) -}}
    {%- elif str(accessor.bool) %}
    {{- bool_template(accessor, name)|indent(4,True) -}}
    {%- elif str(accessor.string) %}
    {{- string_template(accessor, name)|indent(4,True) -}}
    {%- elif str(accessor.enum) and map %}
    {{- enum_template(accessor, name, field.message_type.fields[1])|indent(4,True) -}}
    {%- elif str(accessor.enum) and not map %}
    {{- enum_template(accessor, name, field)|indent(4,True) -}}
    {%- elif str(accessor.duration) %}
    {{- duration_template(accessor, name, True)|indent(4,True) -}}
    {%- elif str(accessor.timestamp) %}
    {{- timestamp_template(accessor, name, True)|indent(4,True) -}}
    {%- elif str(accessor.message) %}
    {{- message_template(accessor, name, True)|indent(4,True) -}}
    {%- elif str(accessor.any) %}
    {{- any_template(accessor, name, True)|indent(4,True) -}}
    {%- elif str(accessor.message) %}
    {{- message_template(accessor, name, True)|indent(4,True) -}}
    {%- endif %}
    """
    return Template(switcher_tmpl).render(
        accessor=accessor,
        name=name,
        str=str,
        num_template=num_template,
        bool_template=bool_template,
        string_template=string_template,
        enum_template=enum_template,
        duration_template=duration_template,
        timestamp_template=timestamp_template,
        any_template=any_template,
        message_template=message_template,
        field=field,
        map=map,
    )


def repeated_template(option_value, name, field):
    rep_tmpl = """
    {%- if o and o.repeated['min_items'] %}
    if len({{ name }}) < {{ o.repeated['min_items'] }}:
        raise ValidationFailed(\"{{ name }} needs to contain at least {{ o.repeated['min_items'] }} items\")
    {%- endif %}
    {%- if o and o.repeated['max_items'] %}
    if len({{ name }}) > {{ o.repeated['max_items'] }}:
        raise ValidationFailed(\"{{ name }} needs to contain at most {{ o.repeated['max_items'] }} items\")
    {%- endif %}
    {%- if o and o.repeated['unique'] %}
    seen = set()
    for item in {{ name }}:
        if item in seen:
            raise ValidationFailed(\"{{ name }} must contain unique items. %s has been repeated.\" %item)
        else:
            seen.add(item)
    {%- endif %}
    {%- if message_type %}
    for item in {{ name }}:
        {%- if o and o.repeated and o.repeated.items.message.skip %}
        pass
        {% else %}
        validate(item)(item)
        {% endif %}
    {%- endif %}
    {%- if o and str(o.repeated['items']) %}
    for item in {{ name }}:
        {%- set accessor = o.repeated['items'] -%}
        {{ switcher_template(accessor, 'item', field) }}
        pass
    {%- endif %}
    """
    return Template(rep_tmpl).render(
        o=option_value,
        name=name,
        message_type=field.message_type,
        str=str,
        field=field,
        switcher_template=switcher_template,
    )


def is_map(field):
    return (
        field.label == 3
        and field.message_type
        and len(field.message_type.fields) == 2
        and field.message_type.fields[0].name == "key"
        and field.message_type.fields[1].name == "value"
    )


def map_template(option_value, name, field):
    map_tmpl = """
    {%- if o and o.map['min_pairs'] %}
    if len({{ name }}) < {{ o.map['min_pairs'] }}:
        raise ValidationFailed(\"{{ name }} needs to contain at least {{ o.map['min_pairs'] }} items\")
    {%- endif %}
    {%- if o and o.map['max_pairs'] %}
    if len({{ name }}) > {{ o.map['max_pairs'] }}:
        raise ValidationFailed(\"{{ name }} can contain at most {{ o.map['max_pairs'] }} items\")
    {%- endif %}
    {%- if o and o.map['no_sparse'] -%}
    raise UnimplementedException(\"no_sparse validation is not implemented because protobuf maps cannot be sparse in Python\")
    {%- endif %}
    {%- if o and (str(o.map['keys']) or str(o.map['values']))%}
    for key in {{ name }}:
        {%- set keys = o.map['keys'] -%}
        {%- set values = o.map['values'] -%}
        {%- if str(keys.double) %}
        {{- num_template(keys, 'key', keys.double)|indent(4,True) -}}
        {%- elif str(keys.int32) %}
        {{- num_template(keys, 'key', keys.int32)|indent(4,True) -}}
        {%- elif str(keys.int64) %}
        {{- num_template(keys, 'key', keys.int64)|indent(4,True) -}}
        {%- elif str(keys.uint32) %}
        {{- num_template(keys, 'key', keys.uint32)|indent(4,True) -}}
        {%- elif str(keys.uint64) %}
        {{- num_template(keys, 'key', keys.uint64)|indent(4,True) -}}
        {%- elif str(keys.sint32) %}
        {{- num_template(keys, 'key', keys.sint32)|indent(4,True) -}}
        {%- elif str(keys.sint64) %}
        {{- num_template(keys, 'key', keys.sint64)|indent(4,True) -}}
        {%- elif str(keys.fixed32) %}
        {{- num_template(keys, 'key', keys.fixed32)|indent(4,True) -}}
        {%- elif str(keys.fixed64) %}
        {{- num_template(keys, 'key', keys.fixed64)|indent(4,True) -}}
        {%- elif str(keys.sfixed32) %}
        {{- num_template(keys, 'key', keys.sfixed32)|indent(4,True) -}}
        {%- elif str(keys.sfixed64) %}
        {{- num_template(keys, 'key', keys.sfixed64)|indent(4,True) -}}
        {%- elif str(keys.bool) %}
        {{- bool_template(keys, 'key')|indent(4,True) -}}
        {%- elif str(keys.string) %}
        {{- string_template(keys, 'key')|indent(4,True) -}}
        {%- endif %}
        {%- set values = o.map['values'] -%}
        {{ switcher_template(values, name +'[key]', field, True) }}
        pass
    {%- elif field.message_type.fields[1].message_type %}
    for key in {{ name }}:
        validate({{ name }}[key])({{ name }}[key])
    {%- endif %}
    """
    return Template(map_tmpl).render(
        o=option_value,
        name=name,
        message_type=field.message_type,
        str=str,
        field=field,
        switcher_template=switcher_template,
        num_template=num_template,
        string_template=string_template,
        bool_template=bool_template,
    )


def rule_type(field):
    name = "p." + field.name
    if has_validate(field) and field.message_type is None:
        for option_descriptor, option_value in field.GetOptions().ListFields():
            if option_descriptor.full_name == "validate.rules":
                if str(option_value.string):
                    return string_template(option_value, name)
                elif str(option_value.message):
                    return message_template(option_value, field.name)
                elif str(option_value.bool):
                    return bool_template(option_value, name)
                elif str(option_value.float):
                    return num_template(option_value, name, option_value.float)
                elif str(option_value.double):
                    return num_template(option_value, name, option_value.double)
                elif str(option_value.int32):
                    return num_template(option_value, name, option_value.int32)
                elif str(option_value.int64):
                    return num_template(option_value, name, option_value.int64)
                elif str(option_value.uint32):
                    return num_template(option_value, name, option_value.uint32)
                elif str(option_value.uint64):
                    return num_template(option_value, name, option_value.uint64)
                elif str(option_value.sint32):
                    return num_template(option_value, name, option_value.sint32)
                elif str(option_value.sint64):
                    return num_template(option_value, name, option_value.sint64)
                elif str(option_value.fixed32):
                    return num_template(option_value, name, option_value.fixed32)
                elif str(option_value.fixed64):
                    return num_template(option_value, name, option_value.fixed64)
                elif str(option_value.sfixed32):
                    return num_template(option_value, name, option_value.sfixed32)
                elif str(option_value.sfixed64):
                    return num_template(option_value, name, option_value.sfixed64)
                elif str(option_value.enum):
                    return enum_template(option_value, name, field)
                elif str(option_value.bytes):
                    return bytes_template(option_value, name)
                elif str(option_value.repeated):
                    return repeated_template(option_value, name, field)
                elif str(option_value.map):
                    return map_template(option_value, name, field)
                elif str(option_value.required):
                    return required_template(option_value, name)
    if field.message_type:
        for option_descriptor, option_value in field.GetOptions().ListFields():
            if option_descriptor.full_name == "validate.rules":
                if str(option_value.duration):
                    return duration_template(option_value, name)
                elif str(option_value.timestamp):
                    return timestamp_template(option_value, name)
                elif (
                    str(option_value.float)
                    or str(option_value.int32)
                    or str(option_value.int64)
                    or str(option_value.double)
                    or str(option_value.uint32)
                    or str(option_value.uint64)
                    or str(option_value.bool)
                    or str(option_value.string)
                    or str(option_value.bytes)
                ):
                    return wrapper_template(option_value, name)
                elif str(option_value.message) != "":
                    return message_template(option_value, field.name)
                elif str(option_value.any):
                    return any_template(option_value, name)
                elif str(option_value.repeated):
                    return repeated_template(option_value, name, field)
                elif str(option_value.map):
                    return map_template(option_value, name, field)
                elif str(option_value.required):
                    return required_template(option_value, name)
        if field.message_type.full_name.startswith("google.protobuf"):
            return ""
        elif is_map(field):
            return map_template(None, name, field)
        elif field.label == 3:
            return repeated_template(None, name, field)
        else:
            return message_template(None, field.name)
    return ""


def file_template(proto_message):
    file_tmp = """
# Validates {{ p.DESCRIPTOR.name }}
def generate_validate(p):
    {%- for option_descriptor, option_value in p.DESCRIPTOR.GetOptions().ListFields() %}
        {%- if option_descriptor.full_name == "validate.disabled" and option_value %}
    return None
        {%- endif -%}
    {%- endfor -%}
    {%- for oneof in p.DESCRIPTOR.oneofs %}
    present = False
        {%- for field in oneof.fields %}
    if _has_field(p, \"{{ field.name }}\"):
        present = True
        {{ rule_type(field)|indent(4,True) }}
        {%- endfor %}
        {% for option in oneof.GetOptions().ListFields() %}
        {% if option[0].name == 'required' and option[1] %}
    if not present:
        raise ValidationFailed(\"Oneof {{ oneof.name }} is required\")
        {% endif %}
        {% endfor %}
    {%- endfor %}
    {%- for field in p.DESCRIPTOR.fields -%}
        {%- if not field.containing_oneof %}
    {{ rule_type(field) -}}
        {%- endif %}
    {%- endfor %}
    return None"""
    return Template(file_tmp).render(rule_type=rule_type, p=proto_message)


class UnimplementedException(Exception):
    pass


class ValidationFailed(Exception):
    pass



import asyncio
import functools
import logging
from abc import abstractmethod, abstractproperty
from dataclasses import dataclass, field
from timeit import default_timer
from typing import Any, Dict, List, Mapping, Optional, Pattern, Type, TypeVar, Union

import faust
from aioprometheus import Counter, Service, Summary
from aioprometheus.collectors import Histogram
from betterproto import Message
from confluent_kafka.admin import AdminClient, NewTopic
from dacite import from_dict
from dataclasses_avroschema import AvroModel
from faust.models import Model
from faust.serializers import Codec
from faust.streams import Stream
from faust.types.serializers import CodecArg, SchemaT
from faust.types.models import ModelT, ModelArg
from faust.types.topics import Seconds, TopicT
from schema_registry.client import SchemaRegistryClient
from schema_registry.serializers.faust import FaustSerializer

from delphai_utils.config import get_config
from delphai_utils.patches import aiokafka_patch, mode_traceback_patch
from delphai_utils.utils import find_free_port

logger = logging.getLogger(__name__)

mode_traceback_patch()
aiokafka_patch()


@dataclass
class Step:
    name: str
    partitions: int
    output: Optional[str]
    tables: Optional[List[str]] = field(default_factory=lambda: [])


T = TypeVar("T")


class FaustProtobufSerializer(Codec):
    def __init__(self, type_class: T, **kwargs):
        self.type_class = type_class
        super(FaustProtobufSerializer, self).__init__(type_class=type_class, **kwargs)

    def _loads(self, s: bytes) -> T:
        return self.type_class().from_json(s.decode("utf-8"))

    def _dumps(self, s: Message) -> bytes:
        return s.to_json().encode("utf-8")


@dataclass
class Metrics:
    inbound_counter: Optional[Counter] = None
    processed_counter: Optional[Counter] = None
    lag_timer: Optional[Summary] = None


class FaustAgent:
    step: str
    step_config: Step
    agent_id: str
    input_topic: faust.TopicT
    output_topic: faust.TopicT
    input_value_type: Type[ModelT]
    output_value_type: Type[ModelT]
    input_value_serializer: CodecArg
    output_value_serializer: CodecArg
    agent: faust.Agent
    metrics: Metrics
    metrics_server: Service

    def setup_metrics(self):
        self.metrics = Metrics()
        name = self.agent_id.replace(".", "_").replace("-", "_")
        self.metrics.inbound_counter = Counter(
            f"delphai_{name}_inbound_counter",
            "incoming requests",
            const_labels={
                "step": self.step,
                "pipeline": self.agent.app.conf.id,
            },
        )
        self.metrics.processed_counter = Counter(
            f"delphai_{name}_processed_counter",
            "processed requests",
            const_labels={
                "step": self.step,
                "pipeline": self.agent.app.conf.id,
            },
        )
        self.metrics.lag_timer = Histogram(
            f"delphai_{name}_lag_timer",
            "time to process requests",
            const_labels={
                "step": self.step,
                "pipeline": self.agent.app.conf.id,
            },
        )
        self.metrics_server.register(self.metrics.inbound_counter)
        self.metrics_server.register(self.metrics.processed_counter)
        self.metrics_server.register(self.metrics.lag_timer)

    @abstractproperty
    def step(self) -> str:
        pass

    @abstractproperty
    def input_value_type(self) -> Type[ModelT]:
        pass

    @abstractproperty
    def output_value_type(self) -> Type[ModelT]:
        pass

    @abstractproperty
    def input_value_serializer(self) -> CodecArg:
        pass

    @abstractproperty
    def output_value_serializer(self) -> CodecArg:
        pass

    @abstractmethod
    async def process(self, requests: Stream[Type[ModelT]]):
        pass

    async def on_start(self):
        logger.info(f"started agent {self.agent_id}")

    async def after_attach(self):
        pass

    def start_measure(self, labels: Dict[str, str] = {}):
        start = default_timer()
        self.metrics.inbound_counter.inc(labels)
        return start

    def end_measure(self, start: float, labels: Dict[str, str] = {}):
        end = default_timer()
        self.metrics.lag_timer.add(labels, end - start)
        self.metrics.processed_counter.inc(labels)


class FaustApp:
    app: faust.App
    worker: faust.Worker
    faust_config: Dict
    metrics: Service = Service()
    pipeline_name: str

    async def on_startup_finished(self):
        logger.info(f"started app {self.app._conf.id}")

    async def on_start(self):
        logger.info(f"starting app {self.app._conf.id}")

    async def on_shutdown(self):
        logger.info(f"shutting down {self.app._conf.id}")

    def __init__(
        self,
        pipeline_name: str,
        broker: str,
        step: Optional[str] = None,
        schema_registry_url: Optional[str] = None,
    ) -> None:
        self.pipeline_name = pipeline_name
        app_id = self.pipeline_name if step is None else f"{self.pipeline_name}.{step}"
        try:
            faust_config = get_config("faust")
            if faust_config is None:
                self.faust_config = {}
            else:
                self.faust_config = faust_config.get("default", {})
                if step:
                    self.faust_config.update(faust_config.get(step, {}))
        except Exception:
            pass
        self.app = _FaustApp(
            id=app_id,
            broker=broker,
            web_enabled=False,
            **self.faust_config,
            schema_registry_url=schema_registry_url,
        )
        loop = asyncio.get_event_loop()

        self.worker = faust.Worker(
            self.app,
            loop=loop,
            override_logging=False,
        )
        self.worker.on_startup_finished = self.on_startup_finished
        self.worker.on_start = self.on_start
        self.worker.on_shutdown = self.on_shutdown

    def start(self, agents: List[FaustAgent]):
        try:
            broker = str(self.app._conf.broker[0])
            steps = list(map(lambda s: s["name"], get_config("steps")))
            for step in steps:
                self.create_topic(broker, step)
            tasks = []
            for agent in agents:
                tasks.append(self.attach(agent))
            metrics_port = find_free_port(9191)
            asyncio.get_event_loop().create_task(self.metrics.start(port=metrics_port))
            logger.info(f"started metrics server on port {metrics_port}")
            self.worker.loop.run_until_complete(asyncio.gather(*tasks))
            self.worker.execute_from_commandline()
        except KeyboardInterrupt:
            logger.info("keyboard interrupt received")

    def get_step_config(self, step: str) -> Step:
        step_config = next((s for s in get_config("steps") if s["name"] == step), None)
        return from_dict(data_class=Step, data=step_config)

    def create_topic(self, broker: str, step: str):
        step_config = self.get_step_config(step)
        client = AdminClient({"bootstrap.servers": broker.replace("kafka://", "")})
        topics = client.list_topics().topics
        topic_names = [f"{self.pipeline_name}.{step_config.name}"]
        for table in step_config.tables:
            table_topic_name = f"{self.pipeline_name}-{table}-changelog"
            topic_names.append(table_topic_name)
        for topic_name in topic_names:
            if topic_name not in topics:
                logger.info(
                    f"creating topic {topic_name} with {step_config.partitions} partitions"
                )
                resp = client.create_topics(
                    [NewTopic(topic_name, step_config.partitions, replication_factor=1)]
                )
                resp[topic_name].result()
            else:
                logger.info(f"topic {topic_name} already exists")

    async def attach(self, agent: FaustAgent):
        agent.step_config = self.get_step_config(agent.step)
        agent.agent_id = f"{self.pipeline_name}.{agent.step}"
        agent.input_topic = self.app.topic(
            agent.agent_id,
            value_type=agent.input_value_type,
            value_serializer=agent.input_value_serializer,
        )
        output_topic_name = f"{self.pipeline_name}.{agent.step_config.output}"
        agent.output_topic = self.app.topic(
            output_topic_name,
            value_serializer=agent.output_value_serializer,
        )
        new_agent = self.app.agent(agent.input_topic)(agent.process)
        agent.agent = new_agent
        agent.metrics_server = self.metrics
        agent.setup_metrics()
        await agent.after_attach()
        await new_agent.start()


def get_avro_schema(cls):
    @functools.wraps(cls, updated=[])
    class WithAvroModel(dataclass(cls), AvroModel):
        pass

    return WithAvroModel.avro_schema()


class _FaustApp(faust.App):
    def __init__(self, *args, schema_registry_url=None, **kwargs):
        self._delphai_schema_registry_client = (
            SchemaRegistryClient(schema_registry_url) if schema_registry_url else None
        )
        super().__init__(*args, **kwargs)

    def topic(
        self,
        *topics: str,
        pattern: Union[str, Pattern] = None,
        schema: SchemaT = None,
        key_type: ModelArg = None,
        value_type: ModelArg = None,
        key_serializer: CodecArg = None,
        value_serializer: CodecArg = None,
        partitions: int = None,
        retention: Seconds = None,
        compacting: bool = None,
        deleting: bool = None,
        replicas: int = None,
        acks: bool = True,
        internal: bool = False,
        config: Mapping[str, Any] = None,
        maxsize: int = None,
        allow_empty: bool = False,
        has_prefix: bool = False,
        loop: asyncio.AbstractEventLoop = None,
    ) -> TopicT:
        if self._delphai_schema_registry_client and (schema is None) and topics:
            if (key_serializer is None) and key_type and issubclass(key_type, Model):
                key_serializer = FaustSerializer(
                    self._delphai_schema_registry_client,
                    f"{topics[0]}-key",
                    get_avro_schema(key_type),
                )

            if (
                (value_serializer is None)
                and value_type
                and issubclass(value_type, Model)
            ):
                value_serializer = FaustSerializer(
                    self._delphai_schema_registry_client,
                    f"{topics[0]}-value",
                    get_avro_schema(value_type),
                )

        return super().topic(
            *topics,
            pattern=pattern,
            schema=schema,
            key_type=key_type,
            value_type=value_type,
            key_serializer=key_serializer,
            value_serializer=value_serializer,
            partitions=partitions,
            retention=retention,
            compacting=compacting,
            deleting=deleting,
            replicas=replicas,
            acks=acks,
            internal=internal,
            config=config,
            maxsize=maxsize,
            allow_empty=allow_empty,
            has_prefix=has_prefix,
            loop=loop,
        )



from multiprocessing import current_process
from threading import Thread
from time import perf_counter, sleep
from abc import ABC, abstractmethod


class Cacher(ABC):
    """
    DEPRECTED! Use redis
    Implements an autorenewable cache for some data
    """

    name = "Abstract cache"
    renew_interval = None
    timeout = 60
    _data = None

    def __init__(self):
        Thread(target=cache_renewer_worker, daemon=True, args=[self]).start()

    def get_data(self):
        """
        Returnes data from cache.
        """
        start = perf_counter()
        while self._data is None:
            if perf_counter() - start > self.timeout:
                raise Exception(f'Data acqisition timeout for cache "{self.name}"')
            sleep(0.2)
        return self._data

    def is_ready(self):
        """Returns True if data acquired and ready to use"""
        return self._data is not None

    @abstractmethod
    def get_data_func(self):
        """Must return something that will be stored into _data"""
        pass


def cache_renewer_worker(cacher: Cacher):
    cacher._data = cacher.get_data_func()
    if cacher.renew_interval:
        while True:
            sleep(cacher.renew_interval)
            cacher._data = cacher.get_data_func()


CACHES_BY_PID = {}  # process id -> cache name -> Cacher

REGISTERED_CACHERS = {}  # cache name -> (Cacher class, autoload)


def register_cacher(cacher_class, autoload: bool = False):
    """DEPRECTED! Use redis"""
    name = cacher_class.name
    REGISTERED_CACHERS[name] = (cacher_class, autoload)
    if autoload:
        get_cache(name)


def get_cache(name) -> Cacher:
    """DEPRECTED! Use redis"""
    pid = current_process().pid
    if pid not in CACHES_BY_PID:
        names2init = set(k for k, v in REGISTERED_CACHERS.items() if v[1])
        names2init.add(name)
        CACHES_BY_PID[pid] = {n: REGISTERED_CACHERS[n][0]() for n in names2init}
    elif name not in CACHES_BY_PID[pid]:
        CACHES_BY_PID[pid][name] = REGISTERED_CACHERS[name][0]()

    return CACHES_BY_PID[pid][name]



import asyncio
import logging
import os
import signal
import socket
import sys
import threading
from functools import partial, wraps

logger = logging.getLogger(__name__)


# Disable
def block_print():
    sys.stdout = open(os.devnull, "w")


# Restore
def enable_print():
    sys.stdout = sys.__stdout__


def _signal_handler(sig, frame):
    logger.info("stopping server...")
    sys.exit(0)


def wait_for_exit():
    signal.signal(signal.SIGINT, _signal_handler)
    forever = threading.Event()
    forever.wait()


def async_wrap(func):
    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        pfunc = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, pfunc)

    return run


def generate_default_message(message_descriptor, lvls: int = 10):
    """
    Returns descriptive template for message.
    """
    if lvls == 0:
        return "..."
    res = {}
    for fld in message_descriptor.fields:
        if fld.message_type:
            fld_content = generate_default_message(fld.message_type, lvls - 1)
        else:
            fld_content = fld.default_value

        if fld.default_value == []:
            if fld.message_type:
                fld_content = [
                    fld_content,
                ]
            elif fld.type == 9:  # string
                fld_content = [
                    "",
                ]
            elif fld.type == 2:  # float
                fld_content = [
                    0.0,
                ]
            elif fld.type == 8:  # boolean
                fld_content = [
                    False,
                ]
            elif fld.type in [3, 5]:  # int32, int64
                fld_content = [
                    0,
                ]
        res[fld.name] = fld_content
    return res


def _is_port_free(host, port):
    """
    determine whether `host` has the `port` free

    From: https://www.thepythoncode.com/article/make-port-scanner-python
    """
    s = socket.socket()
    try:
        s.connect((host, port))
    except Exception:
        return True
    else:
        return False


class NoPortFoundError(Exception):
    ...


def find_free_port(start: int, host="127.0.0.1", num_tries=4) -> int:
    for port in range(start, start + num_tries):
        if _is_port_free(host, port):
            return port
        else:
            logger.info(f"Port {port} already in use.")
    message = f"No free port found in range [{start}, {start + num_tries - 1}]"
    raise NoPortFoundError(message)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['delphai_utils', 'delphai_utils.interceptors']

package_data = \
{'': ['*']}

install_requires = \
['nest-asyncio>=1.5.1,<2.0.0', 'protobuf>=3.20,<4.0']

extras_require = \
{':extra == "streaming"': ['aiokafka==0.7.2',
                           'dataclasses-avroschema>=0.30.3,<0.31.0',
                           'python-schema-registry-client>=2.4.1,<3.0.0'],
 'config': ['omegaconf>=2.1.0,<3.0.0',
            'memoization>=0.3.1,<0.4.0',
            'python-dotenv>=0.14.0,<0.15.0',
            'kubernetes>=25.3.0,<26.0.0',
            'coloredlogs>=14.0,<15.0',
            'deepmerge>=0.1.0,<0.2.0',
            'keyring>=21.5.0,<22.0.0',
            'dacite>=1.6.0,<2.0.0'],
 'database': ['motor>=2.3.0,<3.0.0'],
 'elasticsearch': ['elasticsearch[async]>=7.9.1,<8.0.0'],
 'grpc': ['grpcio>=1.32.0,<2.0.0',
          'starlette>=0,<1',
          'hypercorn>=0.10.2,<0.11.0',
          'validate_email>=1.3,<2.0',
          'ipaddress>=1.0.23,<2.0.0',
          'jinja2>=2.11.2,<3.0.0',
          'markupsafe>=2.0,<2.1',
          'starlette-prometheus>=0.7.0,<0.8.0',
          'grpcio-health-checking>=1.32.0,<2.0.0',
          'grpcio-reflection>=1.32.0,<2.0.0',
          'googleapis-common-protos>=1.52.0,<2.0.0',
          'httpx[http2]>=0.21.0,<0.22.0',
          'python-jose>=3.2.0,<4.0.0',
          'aioprometheus[aiohttp]==21.8.0'],
 'keycloak': ['httpx[http2]>=0.21.0,<0.22.0', 'python-jose>=3.2.0,<4.0.0'],
 'streaming': ['markupsafe>=2.0,<2.1',
               'grpclib>=0.4.2,<0.5.0',
               'faust-streaming>=0.6.9,<0.7.0',
               'confluent-kafka>=1.7.0,<2.0.0',
               'betterproto[compiler]==2.0.0b3',
               'aioprometheus[aiohttp]==21.8.0']}

setup_kwargs = {
    'name': 'delphai-utils',
    'version': '3.0.11',
    'description': 'delphai backend utilities',
    'long_description': None,
    'author': 'Barath Kumar',
    'author_email': 'barath@delphai.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/delphai/delphai-utils',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import time
from typing import Any, Optional, Sequence, Tuple
from grpc.aio import insecure_channel as async_insecure_channel
from grpc import insecure_channel
from random import randint
import grpc


def get_grpc_client(
    Stub,
    address: str,
    is_async: bool = True,
    options: Optional[Sequence[Tuple[str, Any]]] = {},
    max_retries=4,
    use_retry_interceptor=True,
):
    sync_interceptors = (
        SyncUnaryUnaryRetryOnRpcErrorClientInterceptor(
            max_attempts=max_retries,
            sleeping_policy=ExponentialBackoff(
                init_backoff_ms=100, max_backoff_ms=1600, multiplier=2
            ),
            status_for_retry=(grpc.StatusCode.UNAVAILABLE,),
        ),
        SyncUnaryStreamRetryOnRpcErrorClientInterceptor(
            max_attempts=max_retries,
            sleeping_policy=ExponentialBackoff(
                init_backoff_ms=100, max_backoff_ms=1600, multiplier=2
            ),
            status_for_retry=(grpc.StatusCode.UNAVAILABLE,),
        ),
    )
    async_interceptors = (
        AsyncUnaryUnaryRetryOnRpcErrorClientInterceptor(
            max_attempts=max_retries,
            sleeping_policy=ExponentialBackoff(
                init_backoff_ms=100, max_backoff_ms=1600, multiplier=2
            ),
            status_for_retry=(grpc.StatusCode.UNAVAILABLE,),
        ),
    )
    if is_async:
        if use_retry_interceptor:
            channel = async_insecure_channel(address, interceptors=async_interceptors)
        else:
            channel = async_insecure_channel(address, options=options)
    else:
        channel = insecure_channel(address, options=options)
        if use_retry_interceptor:
            channel = grpc.intercept_channel(channel, *sync_interceptors)
    client = Stub(channel)
    return client


# Taken from https://github.com/grpc/grpc/issues/19514#issuecomment-531700657
class ExponentialBackoff:
    def __init__(self, *, init_backoff_ms: int, max_backoff_ms: int, multiplier: int):
        self.init_backoff = randint(0, init_backoff_ms)
        self.max_backoff = max_backoff_ms
        self.multiplier = multiplier

    def sleep(self, try_i: int):
        sleep_range = min(
            self.init_backoff * self.multiplier ** try_i, self.max_backoff
        )
        sleep_ms = randint(0, sleep_range)
        time.sleep(sleep_ms / 1000)


class RetryOnRpcErrorClientInterceptor(
    grpc.UnaryUnaryClientInterceptor, grpc.StreamUnaryClientInterceptor
):
    def __init__(
        self,
        *,
        max_attempts: int,
        sleeping_policy: ExponentialBackoff,
        status_for_retry: Optional[Tuple[grpc.StatusCode]] = None,
    ):
        self.max_attempts = max_attempts
        self.sleeping_policy = sleeping_policy
        self.status_for_retry = status_for_retry

    def _intercept_call(self, continuation, client_call_details, request_or_iterator):

        for try_i in range(self.max_attempts):
            response = continuation(client_call_details, request_or_iterator)

            if isinstance(response, grpc.RpcError):

                is_last_attempt = try_i == (self.max_attempts - 1)
                if is_last_attempt:
                    return response

                is_not_retryable_status = (
                    self.status_for_retry
                    and response.code() not in self.status_for_retry
                )
                if is_not_retryable_status:
                    return response

                self.sleeping_policy.sleep(try_i)
            else:
                return response

    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_unary(
        self, continuation, client_call_details, request_iterator
    ):
        return self._intercept_call(continuation, client_call_details, request_iterator)


class SyncUnaryUnaryRetryOnRpcErrorClientInterceptor(
    RetryOnRpcErrorClientInterceptor, grpc.UnaryUnaryClientInterceptor
):
    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)


class SyncUnaryStreamRetryOnRpcErrorClientInterceptor(
    RetryOnRpcErrorClientInterceptor, grpc.UnaryStreamClientInterceptor
):
    def intercept_unary_stream(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)


class AsyncUnaryUnaryRetryOnRpcErrorClientInterceptor(
    RetryOnRpcErrorClientInterceptor, grpc.aio.UnaryUnaryClientInterceptor
):
    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)



import asyncio
import json
import logging
import os
from time import perf_counter
from typing import Optional
from urllib.parse import urlparse

from google.protobuf import symbol_database
from google.protobuf.json_format import MessageToDict
from grpc import Server, StatusCode
from grpc.experimental.aio import AioRpcError, insecure_channel
from hypercorn.asyncio import serve
from hypercorn.config import Config
from starlette.applications import Starlette
from starlette.exceptions import HTTPException
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette_prometheus import PrometheusMiddleware

from delphai_utils.utils import find_free_port, generate_default_message

logger = logging.getLogger(__name__)

SUPPORTED_METHODS = ["get", "put", "post", "delete", "patch"]
STATUS_MAP = {
    StatusCode.OK: 200,
    StatusCode.CANCELLED: 499,
    StatusCode.UNKNOWN: 500,
    StatusCode.INVALID_ARGUMENT: 400,
    StatusCode.DEADLINE_EXCEEDED: 504,
    StatusCode.NOT_FOUND: 404,
    StatusCode.ALREADY_EXISTS: 409,
    StatusCode.PERMISSION_DENIED: 403,
    StatusCode.UNAUTHENTICATED: 401,
    StatusCode.RESOURCE_EXHAUSTED: 429,
    StatusCode.FAILED_PRECONDITION: 412,
    StatusCode.ABORTED: 499,
    StatusCode.OUT_OF_RANGE: 416,
    StatusCode.UNIMPLEMENTED: 501,
    StatusCode.INTERNAL: 500,
    StatusCode.UNAVAILABLE: 503,
    StatusCode.DATA_LOSS: 420,
}
status_map = STATUS_MAP  # for backward compatibility, don't use it


class AccessLogMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        start = perf_counter()
        response = await call_next(request)
        if response.status_code < 400:
            path = urlparse(str(request.url)).path
            end = perf_counter()
            elapsed = round((end - start) * 1000, 2)
            logger.info(
                f"[{response.status_code}] {request.method} {path} [{elapsed}ms]"
            )
        return response


async def http_exception(request, exc):
    if "favicon.ico" in str(request.url):
        detail = "not found"
        status_code = 404
    else:
        path = urlparse(str(request.url)).path
        logger.error(f"[{exc.status_code}] {request.method} {path} - {exc.detail}")
        detail = exc.detail
        status_code = exc.status_code
    return JSONResponse(
        {"detail": detail, "status": exc.status_code}, status_code=status_code
    )


def _add_method_handlers(method_descriptor, app, channel):
    logger.info(f"  processing {method_descriptor.name}")

    async def method_get_handler(request: Request):
        return JSONResponse(
            {
                "function_name": method_descriptor.full_name,
                "input": generate_default_message(method_descriptor.input_type),
                "output": generate_default_message(method_descriptor.output_type),
            }
        )

    async def method_post_handler(request: Request):
        metadata = request.headers.items()

        body = await request.body()
        body_params = json.loads(body) if body else {}

        input = {**request.path_params, **request.query_params, **body_params}

        try:
            input_prototype = symbol_database.Default().GetPrototype(
                method_descriptor.input_type
            )
            output_prototype = symbol_database.Default().GetPrototype(
                method_descriptor.output_type
            )
            method_callable = channel.unary_unary(
                f"/{method_descriptor.containing_service.full_name}/{method_descriptor.name}",
                request_serializer=input_prototype.SerializeToString,
                response_deserializer=output_prototype.FromString,
            )
            response = await method_callable(
                input_prototype(**input), metadata=metadata
            )

            output = MessageToDict(
                response,
                preserving_proto_field_name=True,
                including_default_value_fields=True,
            )
            return JSONResponse(output)

        except AioRpcError as ex:
            detail = ex.details()
            grpc_status = ex.code()
            http_status_code = STATUS_MAP[grpc_status]
            raise HTTPException(http_status_code, detail=detail)

        except Exception as ex:
            detail = str(ex).replace("\n", " ")
            http_status_code = 500
            raise HTTPException(http_status_code, detail=detail)

    app.add_route(
        f"/{method_descriptor.full_name}",
        route=method_get_handler,
        methods=["get"],
    )
    app.add_route(
        f"/{method_descriptor.full_name}",
        route=method_post_handler,
        methods=["post"],
    )

    for field_descriptor, http_rule in method_descriptor.GetOptions().ListFields():
        if field_descriptor.full_name != "google.api.http":
            continue

        for method in SUPPORTED_METHODS:
            http_path = getattr(http_rule, method)
            if http_path:
                app.add_route(
                    http_path,
                    route=method_post_handler,
                    methods=[method],
                )


def start_gateway(
    server: Server,
    grpc_port: int,
    http_port: Optional[int] = None,
    app: Optional[Starlette] = None,
):
    if not app:
        debug = os.environ.get("DELPHAI_ENVIRONMENT") == "development"
        app = Starlette(debug=debug)

    app.add_exception_handler(HTTPException, http_exception)
    app.add_middleware(AccessLogMiddleware)
    app.add_middleware(PrometheusMiddleware)
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_methods=["*"],
        allow_headers=["*"],
        allow_credentials=True,
    )

    from .grpc_server import GRPC_OPTIONS

    channel = insecure_channel(f"localhost:{grpc_port}", options=GRPC_OPTIONS)

    if not hasattr(server, "descriptors"):
        raise RuntimeError(
            "Server instance does not include the proto file descriptors. Make sure you instantiate it with 'create_grpc_server'"
        )
    for file_descriptor in server.descriptors:
        for service_descriptor in file_descriptor.services_by_name.values():
            if service_descriptor.full_name.startswith("grpc."):
                logger.info(f"skipping service {service_descriptor.name}")
                continue

            logger.info(f"processing service {service_descriptor.name}")
            for method_descriptor in service_descriptor.methods_by_name.values():
                logger.info(f"  processing {method_descriptor.name}")
                _add_method_handlers(method_descriptor, app, channel)

    if not http_port:
        http_port = find_free_port(7070)

    logger.info(f"starting gateway on port {http_port}")
    config = Config()
    config.bind = [f"0.0.0.0:{http_port}"]

    return asyncio.get_event_loop().create_task(serve(app, config))
